package pkg

import (
	"context"
	"github.com/jackc/pgx/v4"
	"github.com/julienschmidt/httprouter"
	"html/template"
	"math"
	"net/http"
)

type ProductEntry struct {
	Slug  string
	Name  string
	Price int
}

type productsIndexPage struct {
	Products []*ProductEntry
}

func IndexProducts(w http.ResponseWriter, r *http.Request, _ httprouter.Params) {
	conn := getConn(r)
	company := mustGetCompany(r)
	page := &productsIndexPage{
		Products: mustGetProductEntries(r.Context(), conn, company),
	}
	mustRenderAppTemplate(w, r, "products/index.gohtml", page)
}

func GetProductForm(w http.ResponseWriter, r *http.Request, params httprouter.Params) {
	locale := getLocale(r)
	conn := getConn(r)
	company := mustGetCompany(r)
	form := newProductForm(r.Context(), conn, locale, company)
	slug := params[0].Value
	if slug == "new" {
		w.WriteHeader(http.StatusOK)
		mustRenderNewProductForm(w, r, form)
		return
	}
	err := conn.QueryRow(r.Context(), "select name, description, price, tax_id from product where slug = $1", slug).Scan(form.Name, form.Description, form.Price, form.Tax)
	if err != nil {
		if err == pgx.ErrNoRows {
			http.NotFound(w, r)
			return
		} else {
			panic(err)
		}
	}
	w.WriteHeader(http.StatusOK)
	mustRenderEditProductForm(w, r, form)
}

func mustRenderNewProductForm(w http.ResponseWriter, r *http.Request, form *productForm) {
	mustRenderAppTemplate(w, r, "products/new.gohtml", form)
}

func mustRenderEditProductForm(w http.ResponseWriter, r *http.Request, form *productForm) {
	mustRenderAppTemplate(w, r, "products/edit.gohtml", form)
}

func HandleAddProduct(w http.ResponseWriter, r *http.Request, _ httprouter.Params) {
	conn := getConn(r)
	locale := getLocale(r)
	company := mustGetCompany(r)
	form := newProductForm(r.Context(), conn, locale, company)
	if err := form.Parse(r); err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}
	if err := verifyCsrfTokenValid(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	if !form.Validate() {
		mustRenderNewProductForm(w, r, form)
		return
	}
	conn.MustExec(r.Context(), "insert into product (company_id, name, description, price, tax_id) values ($1, $2, $3, $4, $5)", company.Id, form.Name, form.Description, form.Price, form.Tax)
	http.Redirect(w, r, companyURI(company, "/products"), http.StatusSeeOther)
}

func HandleUpdateProduct(w http.ResponseWriter, r *http.Request, params httprouter.Params) {
	conn := getConn(r)
	locale := getLocale(r)
	company := mustGetCompany(r)
	form := newProductForm(r.Context(), conn, locale, company)
	if err := form.Parse(r); err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}
	if err := verifyCsrfTokenValid(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	if !form.Validate() {
		mustRenderEditProductForm(w, r, form)
		return
	}
	slug := conn.MustGetText(r.Context(), "", "update product set name = $1, description = $2, price = $3, tax_id = $4 where slug = $5 returning slug", form.Name, form.Description, form.Price, form.Tax, params[0].Value)
	if slug == "" {
		http.NotFound(w, r)
	}
	http.Redirect(w, r, companyURI(company, "/products/"+slug), http.StatusSeeOther)
}

func mustGetProductEntries(ctx context.Context, conn *Conn, company *Company) []*ProductEntry {
	rows, err := conn.Query(ctx, "select slug, name, price from product where company_id = $1 order by name", company.Id)
	if err != nil {
		panic(err)
	}
	defer rows.Close()

	var entries []*ProductEntry
	for rows.Next() {
		entry := &ProductEntry{}
		err = rows.Scan(&entry.Slug, &entry.Name, &entry.Price)
		if err != nil {
			panic(err)
		}
		entries = append(entries, entry)
	}
	if rows.Err() != nil {
		panic(rows.Err())
	}

	return entries
}

type productForm struct {
	locale      *Locale
	Name        *InputField
	Description *InputField
	Price       *InputField
	Tax         *SelectField
}

func newProductForm(ctx context.Context, conn *Conn, locale *Locale, company *Company) *productForm {
	return &productForm{
		locale: locale,
		Name: &InputField{
			Name:     "name",
			Label:    pgettext("input", "Name", locale),
			Type:     "text",
			Required: true,
		},
		Description: &InputField{
			Name:  "description",
			Label: pgettext("input", "Description", locale),
			Type:  "text",
		},
		Price: &InputField{
			Name:     "price",
			Label:    pgettext("input", "Price", locale),
			Type:     "number",
			Required: true,
			Attributes: []template.HTMLAttr{
				`min="0"`,
			},
		},
		Tax: &SelectField{
			Name:    "tax",
			Label:   pgettext("input", "Tax", locale),
			Options: MustGetOptions(ctx, conn, "select tax_id::text, name from tax where company_id = $1 order by name", company.Id),
		},
	}
}

func (form *productForm) Parse(r *http.Request) error {
	if err := r.ParseForm(); err != nil {
		return err
	}
	form.Name.FillValue(r)
	form.Description.FillValue(r)
	form.Price.FillValue(r)
	form.Tax.FillValue(r)
	return nil
}

func (form *productForm) Validate() bool {
	validator := newFormValidator()
	validator.CheckRequiredInput(form.Name, gettext("Name can not be empty.", form.locale))
	if validator.CheckRequiredInput(form.Price, gettext("Price can not be empty.", form.locale)) {
		validator.CheckValidInteger(form.Price, 0, math.MaxInt, gettext("Price must be a number greater than zero.", form.locale))
	}
	validator.CheckValidSelectOption(form.Tax, gettext("Selected tax is not valid.", form.locale))
	return validator.AllOK()
}
